@@ -3,6 +3,7 @@
  * Copyright (C) 2009-2011 Vic Lee
  * Copyright (C) 2014-2015 Antenore Gatta, Fabio Castelli, Giovanni Panozzo
  * Copyright (C) 2016-2023 Antenore Gatta, Giovanni Panozzo
+ * Copyright (C) 2023-2024 Hiroyuki Tanaka, Sunil Bhat
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -111,16 +112,24 @@ struct _RemminaFileEditorPriv {
 	GtkWidget *		resolution_custom_combo;
 	GtkWidget *		keymap_combo;
 
+	GtkWidget *		assistance_toggle;
+	GtkWidget *		assistance_file;
+	GtkWidget *		assistance_password;
+	GtkWidget *		assistance_file_label;
+	GtkWidget *		assistance_password_label;
+
 	GtkWidget *		behavior_autostart_check;
 	GtkWidget *		behavior_precommand_entry;
 	GtkWidget *		behavior_postcommand_entry;
 	GtkWidget *		behavior_lock_check;
+	GtkWidget *		behavior_disconnect;
 
 	GtkWidget *		ssh_tunnel_enabled_check;
 	GtkWidget *		ssh_tunnel_loopback_check;
 	GtkWidget *		ssh_tunnel_server_default_radio;
 	GtkWidget *		ssh_tunnel_server_custom_radio;
 	GtkWidget *		ssh_tunnel_server_entry;
+	GtkWidget *		ssh_tunnel_command_entry;
 	GtkWidget *		ssh_tunnel_auth_agent_radio;
 	GtkWidget *		ssh_tunnel_auth_password_radio;
 	GtkWidget *		ssh_tunnel_auth_password;
@@ -354,6 +363,26 @@ static GtkWidget *remmina_file_editor_create_notebook_tab(RemminaFileEditor *gfe
 	return grid;
 }
 
+
+static void remmina_file_editor_assistance_enabled_check_on_toggled(GtkToggleButton *togglebutton,
+								    RemminaFileEditor *gfe)
+{
+	TRACE_CALL(__func__);
+	gboolean enabled = TRUE;
+
+	if (gfe->priv->assistance_toggle) {
+		enabled = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gfe->priv->assistance_toggle));
+		if (gfe->priv->assistance_file)
+			gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->assistance_file), enabled);
+		if (gfe->priv->assistance_password)
+			gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->assistance_password), enabled);
+		if (gfe->priv->assistance_file_label)
+			gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->assistance_file_label), enabled);
+		if (gfe->priv->assistance_password_label)
+			gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->assistance_password_label), enabled);
+	}
+}
+
 #ifdef HAVE_LIBSSH
 
 static void remmina_file_editor_ssh_tunnel_server_custom_radio_on_toggled(GtkToggleButton *togglebutton, RemminaFileEditor *gfe)
@@ -365,6 +394,7 @@ static void remmina_file_editor_ssh_tunnel_server_custom_radio_on_toggled(GtkTog
 				  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(gfe->priv->ssh_tunnel_server_custom_radio))));
 }
 
+
 static void remmina_file_editor_ssh_tunnel_enabled_check_on_toggled(GtkToggleButton *togglebutton,
 								    RemminaFileEditor *gfe, RemminaProtocolSSHSetting ssh_setting)
 {
@@ -385,16 +415,13 @@ static void remmina_file_editor_ssh_tunnel_enabled_check_on_toggled(GtkToggleBut
 			gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->ssh_tunnel_server_custom_radio), enabled);
 		remmina_file_editor_ssh_tunnel_server_custom_radio_on_toggled(NULL, gfe);
 		p = remmina_public_combo_get_active_text(GTK_COMBO_BOX(priv->protocol_combo));
-		// if (!(g_strcmp0(p, "SFTP") == 0 || g_strcmp0(p, "SSH") == 0)) {
 		gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->ssh_tunnel_username_entry), enabled);
 		gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->ssh_tunnel_auth_password), enabled);
 		gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->ssh_tunnel_auth_combo), enabled);
-		gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->ssh_tunnel_privatekey_chooser), enabled);
-		gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->ssh_tunnel_certfile_chooser), enabled);
-		//}
+		gtk_widget_set_sensitive(GTK_WIDGET(gfe->priv->ssh_tunnel_command_entry), enabled);
+		
 		g_free(p);
 	}
-	// remmina_file_editor_ssh_tunnel_auth_publickey_radio_on_toggled(NULL, gfe);
 	s = remmina_file_get_string(gfe->priv->remmina_file, "ssh_tunnel_privatekey");
 	if (s)
 		gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(gfe->priv->ssh_tunnel_privatekey_chooser), s);
@@ -420,6 +447,12 @@ static void remmina_file_editor_ssh_tunnel_enabled_check_on_toggled(GtkToggleBut
 			gtk_entry_set_text(GTK_ENTRY(gfe->priv->ssh_tunnel_passphrase), cp ? cp : "");
 		}
 	}
+	if (gfe->priv->ssh_tunnel_command_entry) {
+		if (enabled && gtk_entry_get_text(GTK_ENTRY(gfe->priv->ssh_tunnel_command_entry))[0] == '\0') {
+			cp = remmina_file_get_string(priv->remmina_file, "ssh_tunnel_command");
+			gtk_entry_set_text(GTK_ENTRY(gfe->priv->ssh_tunnel_command_entry), cp ? cp : "");
+		}
+	}
 }
 
 #endif
@@ -473,23 +506,8 @@ static void remmina_file_editor_create_server(RemminaFileEditor *gfe, const Remm
 		gtk_grid_attach(GTK_GRID(grid), widget, 1, row, 1, 1);
 	}
 }
-static void remmina_file_toggle_password_view(GtkWidget *widget, gpointer data)
-{
-	GtkWindow *mainwindow;
-	gboolean visible = gtk_entry_get_visibility(GTK_ENTRY(widget));
 
-	mainwindow = remmina_main_get_window();
-	if (remmina_pref_get_boolean("use_primary_password") && remmina_pref_get_boolean("lock_view_passwords") && remmina_unlock_new(mainwindow) == 0)
-		return;
 
-	if (visible) {
-		gtk_entry_set_visibility(GTK_ENTRY(widget), FALSE);
-		gtk_entry_set_icon_from_icon_name(GTK_ENTRY(widget), GTK_ENTRY_ICON_SECONDARY, "org.remmina.Remmina-password-reveal-symbolic");
-	} else {
-		gtk_entry_set_visibility(GTK_ENTRY(widget), TRUE);
-		gtk_entry_set_icon_from_icon_name(GTK_ENTRY(widget), GTK_ENTRY_ICON_SECONDARY, "org.remmina.Remmina-password-conceal-symbolic");
-	}
-}
 static GtkWidget *remmina_file_editor_create_password(RemminaFileEditor *gfe, GtkWidget *grid, gint row, gint col, const gchar *label, const gchar *value, gchar *setting_name)
 {
 	TRACE_CALL(__func__);
@@ -509,7 +527,7 @@ static GtkWidget *remmina_file_editor_create_password(RemminaFileEditor *gfe, Gt
 	widget = gtk_entry_new();
 	gtk_widget_show(widget);
 	gtk_grid_attach(GTK_GRID(grid), widget, 1, row, 1, 1);
-	gtk_entry_set_max_length(GTK_ENTRY(widget), 100);
+	gtk_entry_set_max_length(GTK_ENTRY(widget), 0);
 	gtk_entry_set_visibility(GTK_ENTRY(widget), FALSE);
 	gtk_widget_set_hexpand(widget, TRUE);
 	gtk_entry_set_activates_default(GTK_ENTRY(widget), TRUE);
@@ -522,7 +540,7 @@ static GtkWidget *remmina_file_editor_create_password(RemminaFileEditor *gfe, Gt
 	if (setting_name) {
 		gtk_entry_set_icon_from_icon_name(GTK_ENTRY(widget), GTK_ENTRY_ICON_SECONDARY, "org.remmina.Remmina-password-reveal-symbolic");
 		gtk_entry_set_icon_activatable(GTK_ENTRY(widget), GTK_ENTRY_ICON_SECONDARY, TRUE);
-		g_signal_connect(widget, "icon-press", G_CALLBACK(remmina_file_toggle_password_view), NULL);
+		g_signal_connect(widget, "icon-press", G_CALLBACK(remmina_main_toggle_password_view), NULL);
 	}
 	return widget;
 }
@@ -634,6 +652,61 @@ static void remmina_file_editor_create_resolution(RemminaFileEditor *gfe, const
 	g_free(res_str);
 }
 
+
+static void remmina_file_editor_create_assistance(RemminaFileEditor *gfe, const RemminaProtocolSetting *setting,
+						  GtkWidget *grid, gint row)
+{
+	TRACE_CALL(__func__);
+	GtkWidget *widget;
+
+
+
+	widget = gtk_toggle_button_new_with_label(_("Assistance Mode"));
+	gtk_widget_set_halign(widget, GTK_ALIGN_START);
+	gtk_widget_show(widget);
+	gtk_grid_attach(GTK_GRID(grid), widget, 0, row, 1, 1);
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), remmina_file_get_int(gfe->priv->remmina_file, "assistance_mode", 0));
+	gfe->priv->assistance_toggle = widget;
+	g_signal_connect(widget, "toggled", G_CALLBACK(remmina_file_editor_assistance_enabled_check_on_toggled), gfe);
+
+
+	widget = gtk_label_new("Assistance file");
+	gtk_widget_set_halign(widget, GTK_ALIGN_END);
+	gtk_widget_show(widget);
+	gtk_grid_attach(GTK_GRID(grid), widget, 0, row+1, 1, 1);
+	gfe->priv->assistance_file_label = widget;
+
+	widget = gtk_entry_new();
+	gtk_widget_set_halign(widget, GTK_ALIGN_START);
+	gtk_widget_show(widget);
+	
+	if (remmina_file_get_string(gfe->priv->remmina_file, "assistance_file") != NULL) {
+		gtk_entry_set_text(GTK_ENTRY(widget), remmina_file_get_string(gfe->priv->remmina_file, "assistance_file"));
+	}
+	gtk_grid_attach(GTK_GRID(grid), widget, 1, row+1, 1, 1);
+	gfe->priv->assistance_file = widget;
+
+	widget = gtk_label_new("Assistance Password");
+	gtk_widget_set_halign(widget, GTK_ALIGN_END);
+	gtk_widget_show(widget);
+	gtk_grid_attach(GTK_GRID(grid), widget, 0, row+2, 1, 1);
+	gfe->priv->assistance_password_label = widget;
+
+	widget = gtk_entry_new();
+	gtk_widget_set_halign(widget, GTK_ALIGN_START);
+	gtk_widget_show(widget);
+
+	if (remmina_file_get_string(gfe->priv->remmina_file, "assistance_pass") != NULL) {
+		gtk_entry_set_text(GTK_ENTRY(widget), remmina_file_get_string(gfe->priv->remmina_file, "assistance_pass"));
+	}
+	gtk_grid_attach(GTK_GRID(grid), widget, 1, row+2, 1, 1);
+	gfe->priv->assistance_password = widget;
+
+	remmina_file_editor_assistance_enabled_check_on_toggled(NULL, gfe);
+
+}
+
+
 static GtkWidget *remmina_file_editor_create_text2(RemminaFileEditor *gfe, GtkWidget *grid,
 						   gint row, gint col, const gchar *label, const gchar *value, gint left,
 						   gint right, gchar *setting_name)
@@ -773,6 +846,240 @@ static GtkWidget *remmina_file_editor_create_check(RemminaFileEditor *gfe, GtkWi
 	return widget;
 }
 
+static void remmina_file_editor_export_settings(RemminaFileEditor* rfe)
+{
+	RemminaFileEditorPriv *priv = rfe->priv;
+}
+
+static void remmina_file_editor_save_ssh_tunnel_tab(RemminaFileEditor *gfe, RemminaFile* ssh_file)
+{
+	TRACE_CALL(__func__);
+	RemminaFileEditorPriv *priv = gfe->priv;
+	gboolean ssh_tunnel_enabled;
+	int ssh_tunnel_auth;
+	RemminaFile* remminafile;
+	if (ssh_file == NULL){
+		remminafile = priv->remmina_file;
+	}
+	else{
+		remminafile = ssh_file;
+	}
+	
+
+	ssh_tunnel_enabled = (priv->ssh_tunnel_enabled_check ? gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_enabled_check)) : FALSE);
+	remmina_file_set_int(remminafile,
+			     "ssh_tunnel_loopback",
+			     (priv->ssh_tunnel_loopback_check ? gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_loopback_check)) : FALSE));
+	remmina_file_set_int(remminafile, "ssh_tunnel_enabled", ssh_tunnel_enabled);
+	remmina_file_set_string(remminafile, "ssh_tunnel_auth",
+				remmina_public_combo_get_active_text(GTK_COMBO_BOX(priv->ssh_tunnel_auth_combo)));
+	remmina_file_set_string(remminafile, "ssh_tunnel_username",
+				(ssh_tunnel_enabled ? gtk_entry_get_text(GTK_ENTRY(priv->ssh_tunnel_username_entry)) : NULL));
+	remmina_file_set_string(
+		remminafile,
+		"ssh_tunnel_server",
+		(ssh_tunnel_enabled && priv->ssh_tunnel_server_entry && (priv->ssh_tunnel_server_custom_radio == NULL || gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_server_custom_radio))) ? gtk_entry_get_text(GTK_ENTRY(priv->ssh_tunnel_server_entry)) : NULL));
+
+	ssh_tunnel_auth = gtk_combo_box_get_active(GTK_COMBO_BOX(priv->ssh_tunnel_auth_combo));
+
+	remmina_file_set_int(
+		remminafile,
+		"ssh_tunnel_auth",
+		ssh_tunnel_auth);
+	
+	// If box is unchecked for private key and certfile file choosers,
+	// set the string to NULL in the remmina file 
+	if (gtk_widget_get_sensitive(priv->ssh_tunnel_privatekey_chooser)) {
+		remmina_file_set_string(
+		remminafile,
+		"ssh_tunnel_privatekey",
+		(priv->ssh_tunnel_privatekey_chooser ? gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(priv->ssh_tunnel_privatekey_chooser)) : NULL));
+	}
+	else {
+		remmina_file_set_string(remminafile, "ssh_tunnel_privatekey", NULL);
+	}
+	
+	if (gtk_widget_get_sensitive(priv->ssh_tunnel_certfile_chooser)) {
+		remmina_file_set_string(
+			remminafile,
+			"ssh_tunnel_certfile",
+			(priv->ssh_tunnel_certfile_chooser ? gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(priv->ssh_tunnel_certfile_chooser)) : NULL));
+	}
+	else {
+		remmina_file_set_string(remminafile, "ssh_tunnel_certfile", NULL);
+	}
+
+	remmina_file_set_string(
+		remminafile,
+		"ssh_tunnel_password",
+		(ssh_tunnel_enabled && (ssh_tunnel_auth == SSH_AUTH_PASSWORD)) ? gtk_entry_get_text(GTK_ENTRY(priv->ssh_tunnel_auth_password)) : NULL);
+
+	const char* command = gtk_entry_get_text(GTK_ENTRY(priv->ssh_tunnel_command_entry));
+	remmina_file_set_string(
+		remminafile,
+		"ssh_tunnel_command", command);
+
+	remmina_file_set_string(
+		remminafile,
+		"ssh_tunnel_passphrase",
+		(ssh_tunnel_enabled && (ssh_tunnel_auth == SSH_AUTH_PUBLICKEY || ssh_tunnel_auth == SSH_AUTH_AUTO_PUBLICKEY)) ? gtk_entry_get_text(GTK_ENTRY(priv->ssh_tunnel_passphrase)) : NULL);
+}
+
+
+#ifdef HAVE_LIBSSH
+static gpointer ssh_tunnel_auth_list[] =
+{
+	"0", N_("Password"),
+	"1", N_("SSH identity file"),
+	"2", N_("SSH agent"),
+	"3", N_("Public key (automatic)"),
+	"4", N_("Kerberos (GSSAPI)"),
+	NULL
+};
+#endif
+
+
+static void remmina_file_editor_run_import_export_dialog(gpointer user_data, gint type)
+{
+
+	RemminaFileEditor* gfe = (RemminaFileEditor*)user_data;
+	RemminaFileEditorPriv *priv = gfe->priv;
+	const gchar *cs;
+	gint ci;
+	GtkWidget* dialog;
+	if (type == GTK_FILE_CHOOSER_ACTION_SAVE){
+		dialog = gtk_file_chooser_dialog_new(_("Save to..."), NULL, GTK_FILE_CHOOSER_ACTION_SAVE, _("_Cancel"),
+		GTK_RESPONSE_CANCEL, _("Save"), GTK_RESPONSE_ACCEPT, NULL);
+	}
+	else{
+		dialog = gtk_file_chooser_dialog_new(_("Open ssh tunnel config"), NULL, GTK_FILE_CHOOSER_ACTION_SAVE, _("_Cancel"),
+		GTK_RESPONSE_CANCEL, _("Import"), GTK_RESPONSE_ACCEPT, NULL);
+	}
+	
+
+	GtkFileFilter *filter;
+
+
+	filter = gtk_file_filter_new();
+	gtk_file_filter_set_name(filter, _("Remmina Files"));
+	gtk_file_filter_add_pattern(filter, "*.remssh");
+	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);
+	gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(dialog), _("ssh_export.remssh"));
+	gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), remmina_pref.datadir_path);
+	gint response = gtk_dialog_run(GTK_DIALOG(dialog));
+	
+
+	if (response == GTK_RESPONSE_ACCEPT && type == GTK_FILE_CHOOSER_ACTION_SAVE){
+		RemminaFile* export_rf = (RemminaFile*)remmina_file_new();
+		remmina_file_editor_save_ssh_tunnel_tab(gfe, export_rf);
+		
+		export_rf->filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
+		remmina_file_save(export_rf);
+
+	}
+	else if (response == GTK_RESPONSE_ACCEPT && type == GTK_FILE_CHOOSER_ACTION_OPEN){
+		gchar* import_filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
+		RemminaFile* import_rf = remmina_file_load(import_filename);
+		//get tunnel enabled
+		ci = remmina_file_get_int(import_rf, "ssh_tunnel_enabled", 0);
+		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_enabled_check), ci);
+		//get tunnel loopback address
+		ci = remmina_file_get_int(import_rf, "ssh_tunnel_loopback", 0);
+		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_loopback_check), ci);
+		//get the toggle button
+		cs = remmina_file_get_string(import_rf, "ssh_tunnel_server");
+		if (cs != NULL){
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_enabled_check),
+							remmina_file_get_int(import_rf, "ssh_tunnel_enabled", FALSE));
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_loopback_check),
+							remmina_file_get_int(import_rf, "ssh_tunnel_loopback", FALSE));
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cs ? priv->ssh_tunnel_server_custom_radio : priv->ssh_tunnel_server_default_radio), TRUE);
+			gtk_entry_set_text(GTK_ENTRY(priv->ssh_tunnel_server_entry),
+					cs ? cs : "");
+		}
+		//get authentication
+		ci = remmina_file_get_int(import_rf, "ssh_tunnel_auth", 0);
+		gint i = 0;
+		for (i = 0; ssh_tunnel_auth_list[i]; i+=2){
+			if (atoi(ssh_tunnel_auth_list[i]) == ci){
+				gtk_combo_box_set_active(GTK_COMBO_BOX(priv->ssh_tunnel_auth_combo), i / 2);
+			}
+		}
+		//get username
+		cs = remmina_file_get_string(import_rf, "ssh_tunnel_username");
+		if (cs != NULL){
+			gtk_entry_set_text(GTK_ENTRY(priv->ssh_tunnel_username_entry), cs);
+		}
+		g_free(cs);
+		//get password
+		cs = remmina_file_get_string(import_rf, "ssh_tunnel_password");
+		if (cs != NULL){
+			gtk_entry_set_text(GTK_ENTRY(priv->ssh_tunnel_auth_password), cs);
+		}
+		g_free(cs);
+		//get private key
+		cs = remmina_file_get_string(import_rf, "ssh_tunnel_privatekey");
+		if (cs != NULL){
+			GtkWidget* check = GTK_WIDGET(g_object_get_data(G_OBJECT(priv->ssh_tunnel_privatekey_chooser), "check"));
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), TRUE);
+			gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(priv->ssh_tunnel_privatekey_chooser), cs);
+		}
+		g_free(cs);
+		//get certificate
+		cs = remmina_file_get_string(import_rf, "ssh_tunnel_certfile");
+		if (cs != NULL){
+			GtkWidget* check = GTK_WIDGET(g_object_get_data(G_OBJECT(priv->ssh_tunnel_certfile_chooser), "check"));
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), TRUE);
+			gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(priv->ssh_tunnel_certfile_chooser), cs);
+		}
+		g_free(cs);
+		//get key passoword
+		cs = remmina_file_get_string(import_rf, "ssh_tunnel_passphrase");
+		if (cs != NULL){
+			gtk_entry_set_text(GTK_ENTRY(priv->ssh_tunnel_passphrase), cs);
+		}
+		//get startup command
+		cs = remmina_file_get_string(import_rf, "ssh_tunnel_command");
+		if (cs != NULL){
+			gtk_entry_set_text(GTK_ENTRY(priv->ssh_tunnel_command_entry), cs);
+		}
+		g_free(cs);
+	}	
+	gtk_widget_destroy(dialog);
+}
+
+
+static void remmina_file_editor_run_export(GtkButton* self, gpointer user_data){
+	remmina_file_editor_run_import_export_dialog(user_data, GTK_FILE_CHOOSER_ACTION_SAVE);
+}
+
+
+static void remmina_file_editor_run_import(GtkButton* self, gpointer user_data){
+	remmina_file_editor_run_import_export_dialog(user_data, GTK_FILE_CHOOSER_ACTION_OPEN);
+}
+
+
+static GtkWidget* remmina_file_editor_create_sssh_import_export(RemminaFileEditor *gfe, GtkWidget *grid, gint row, gint col, const gchar *label,
+				   const gchar *value, gint type, gchar *setting_name)
+{
+	GtkWidget* widget = gtk_button_new_with_label(label);
+	gtk_widget_set_name(widget, label);
+	gtk_grid_attach(GTK_GRID(grid), widget, 0, row, 1, 1);
+	
+	
+	if (type == GTK_FILE_CHOOSER_ACTION_SAVE){
+		g_signal_connect(G_OBJECT(widget), "clicked", G_CALLBACK(remmina_file_editor_run_export), gfe);
+	}
+	else{
+		g_signal_connect(G_OBJECT(widget), "clicked", G_CALLBACK(remmina_file_editor_run_import), gfe);
+	}
+	
+
+	return widget;
+
+}
+
+
 /**
  * Create checkbox + gtk_file_chooser for open files and select folders
  *
@@ -817,6 +1124,7 @@ remmina_file_editor_create_chooser(RemminaFileEditor *gfe, GtkWidget *grid, gint
 	g_signal_connect(G_OBJECT(check), "toggled", G_CALLBACK(remmina_file_editor_button_on_toggled), widget);
 	remmina_file_editor_button_on_toggled(GTK_TOGGLE_BUTTON(check), widget);
 
+	g_object_set_data(G_OBJECT(widget), "check", (gpointer)check);
 	return widget;
 }
 
@@ -931,7 +1239,7 @@ static GtkWidget *remmina_file_editor_create_double(RemminaFileEditor *gfe,
 	return widget;
 }
 
-gdouble remmina_file_get_double(RemminaFile *remminafile, const gchar *setting, gfloat default_value);
+
 
 static void remmina_file_editor_create_settings(RemminaFileEditor *gfe, GtkWidget *grid,
 						const RemminaProtocolSetting *settings)
@@ -963,7 +1271,12 @@ static void remmina_file_editor_create_settings(RemminaFileEditor *gfe, GtkWidge
 
 		case REMMINA_PROTOCOL_SETTING_TYPE_RESOLUTION:
 			remmina_file_editor_create_resolution(gfe, settings, grid, grid_row);
-			grid_row += 1;
+			grid_row ++;
+			break;
+
+		case REMMINA_PROTOCOL_SETTING_TYPE_ASSISTANCE:
+			remmina_file_editor_create_assistance(gfe, settings, grid, grid_row);
+			grid_row += 3;
 			break;
 
 		case REMMINA_PROTOCOL_SETTING_TYPE_KEYMAP:
@@ -985,6 +1298,9 @@ static void remmina_file_editor_create_settings(RemminaFileEditor *gfe, GtkWidge
 			g_hash_table_insert(priv->setting_widgets, setting_name, widget);
 			if (settings->opt2)
 				gtk_widget_set_tooltip_text(widget, _((const gchar *)settings->opt2));
+			if (settings->opt1){
+				gtk_entry_set_placeholder_text(GTK_ENTRY(widget), _((const gchar *)settings->opt1));
+			}
 			grid_row++;
 			break;
 
@@ -1102,46 +1418,55 @@ static void remmina_file_editor_create_behavior_tab(RemminaFileEditor *gfe)
 	/* Execute Command frame */
 	remmina_public_create_group(GTK_GRID(grid), _("Execute a Command"), 0, 1, 2);
 
+	/* Sandbox disclaimer*/
+	widget = gtk_label_new(NULL);
+	gtk_widget_show(widget);
+	gtk_widget_set_halign(GTK_WIDGET(widget), GTK_ALIGN_START);
+	gtk_widget_set_valign(GTK_WIDGET(widget), GTK_ALIGN_CENTER);
+	gtk_label_set_text(GTK_LABEL(widget), "* Local commands may not function properly if run in a sandboxed environment *");
+	gtk_grid_attach(GTK_GRID(grid), widget, 0, 2, 2, 2);
+
 	/* PRE connection command */
 	cs = remmina_file_get_string(priv->remmina_file, "precommand");
-	widget = remmina_file_editor_create_text2(gfe, grid, 2, 0, _("Before connecting"), cs, 24, 26, "precommand");
+	widget = remmina_file_editor_create_text2(gfe, grid, 4, 0, _("Before connecting"), cs, 24, 26, "precommand");
 	priv->behavior_precommand_entry = widget;
 	gtk_entry_set_placeholder_text(GTK_ENTRY(widget), _("command %h %u %t %U %p %g --option"));
 	gtk_widget_set_tooltip_markup(widget, _(cmd_tips));
 
 	/* POST connection command */
 	cs = remmina_file_get_string(priv->remmina_file, "postcommand");
-	widget = remmina_file_editor_create_text2(gfe, grid, 3, 0, _("After connecting"), cs, 24, 16, "postcommand");
+	widget = remmina_file_editor_create_text2(gfe, grid, 5, 0, _("After connecting"), cs, 24, 16, "postcommand");
 	priv->behavior_postcommand_entry = widget;
 	gtk_entry_set_placeholder_text(GTK_ENTRY(widget), _("/path/to/command -opt1 arg %h %u %t -opt2 %U %p %g"));
 	gtk_widget_set_tooltip_markup(widget, _(cmd_tips));
 
 	/* Startup frame */
-	remmina_public_create_group(GTK_GRID(grid), _("Start-up"), 4, 1, 2);
+	remmina_public_create_group(GTK_GRID(grid), _("Start-up"), 6, 1, 2);
 
 	/* Autostart profile option */
-	priv->behavior_autostart_check = remmina_file_editor_create_check(gfe, grid, 6, 1, _("Auto-start this profile"),
+	priv->behavior_autostart_check = remmina_file_editor_create_check(gfe, grid, 8, 1, _("Auto-start this profile"),
 									  remmina_file_get_int(priv->remmina_file, "enable-autostart", FALSE), "enable-autostart");
 
 	/* Startup frame */
-	remmina_public_create_group(GTK_GRID(grid), _("Connection profile security"), 8, 1, 2);
+	remmina_public_create_group(GTK_GRID(grid), _("Connection profile security"), 10, 1, 2);
 
 	/* Autostart profile option */
-	priv->behavior_lock_check = remmina_file_editor_create_check(gfe, grid, 10, 1, _("Require password to connect or edit the profile"),
+	priv->behavior_lock_check = remmina_file_editor_create_check(gfe, grid, 12, 1, _("Require password to connect or edit the profile"),
 								     remmina_file_get_int(priv->remmina_file, "profile-lock", FALSE), "profile-lock");
+
+									 /* Startup frame */
+	remmina_public_create_group(GTK_GRID(grid), _("Unexpected disconnect"), 14, 1, 2);
+
+	/* Autostart profile option */
+	priv->behavior_disconnect = remmina_file_editor_create_check(gfe, grid, 18, 1, _("Keep window from closing if not disconnected by Remmina"),
+								     remmina_file_get_int(priv->remmina_file, "disconnect-prompt", FALSE), "disconnect-prompt");
 }
 
-#ifdef HAVE_LIBSSH
-static gpointer ssh_tunnel_auth_list[] =
-{
-	"0", N_("Password"),
-	"1", N_("SSH identity file"),
-	"2", N_("SSH agent"),
-	"3", N_("Public key (automatic)"),
-	"4", N_("Kerberos (GSSAPI)"),
-	NULL
-};
-#endif
+
+static gboolean
+remmina_file_editor_ssh_tunnel_import_settings(GtkFileChooserButton* self, gpointer user_data){
+	g_debug("Importing settings from %s", gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(self)));
+}
 
 static void remmina_file_editor_create_ssh_tunnel_tab(RemminaFileEditor *gfe, RemminaProtocolSSHSetting ssh_setting)
 {
@@ -1160,7 +1485,7 @@ static void remmina_file_editor_create_ssh_tunnel_tab(RemminaFileEditor *gfe, Re
 
 	/* The SSH tab (implementation) */
 	grid = remmina_file_editor_create_notebook_tab(gfe, NULL,
-						       _("SSH Tunnel"), 9, 3);
+						       _("SSH Tunnel"), 12, 3);
 	widget = gtk_toggle_button_new_with_label(_("Enable SSH tunnel"));
 	gtk_widget_set_halign(widget, GTK_ALIGN_START);
 	gtk_grid_attach(GTK_GRID(grid), widget, 0, row, 1, 1);
@@ -1305,6 +1630,37 @@ static void remmina_file_editor_create_ssh_tunnel_tab(RemminaFileEditor *gfe, Re
 				   cs ? cs : "");
 	}
 
+
+	remmina_public_create_group(GTK_GRID(grid), _("SSH Command"), row, 2, 1);
+	row += 2;
+	priv->ssh_tunnel_command_entry =
+			remmina_file_editor_create_text(gfe, grid, row, 0,
+							_("Startup command"), NULL, "ssh_tunnel_command");
+	cs = remmina_file_get_string(priv->remmina_file, "ssh_tunnel_command");
+	gtk_entry_set_text(GTK_ENTRY(priv->ssh_tunnel_command_entry),
+				   cs ? cs : "");
+	row++;
+
+	widget = remmina_file_editor_create_sssh_import_export(gfe, grid, row, 0,
+									       _("Export ssh tunnel settings"),
+									       remmina_file_get_string(priv->remmina_file, "ssh_saved_profile"),
+									       GTK_FILE_CHOOSER_ACTION_SAVE, "ssh_export_profile");
+	row++;
+
+	widget = remmina_file_editor_create_sssh_import_export(gfe, grid, row, 1,
+									       _("Import ssh tunnel settings"),
+									       remmina_file_get_string(priv->remmina_file, "ssh_saved_profile"),
+									       GTK_FILE_CHOOSER_ACTION_OPEN, "ssh_import_profile");
+
+	// row++;
+
+
+
+	row++;
+
+
+
+
 	remmina_file_editor_ssh_tunnel_enabled_check_on_toggled(NULL, gfe, ssh_setting);
 	gtk_widget_show_all(grid);
 	g_free(p);
@@ -1376,6 +1732,7 @@ static void remmina_file_editor_protocol_combo_on_changed(GtkComboBox *combo, Re
 	priv->ssh_tunnel_server_custom_radio = NULL;
 	priv->ssh_tunnel_server_entry = NULL;
 	priv->ssh_tunnel_username_entry = NULL;
+	priv->ssh_tunnel_command_entry = NULL;
 	priv->ssh_tunnel_auth_combo = NULL;
 	priv->ssh_tunnel_auth_password = NULL;
 	priv->ssh_tunnel_privatekey_chooser = NULL;
@@ -1404,55 +1761,10 @@ static void remmina_file_editor_save_behavior_tab(RemminaFileEditor *gfe)
 	remmina_file_set_int(priv->remmina_file, "enable-autostart", autostart_enabled);
 	gboolean lock_enabled = (priv->behavior_lock_check ? gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->behavior_lock_check)) : FALSE);
 	remmina_file_set_int(priv->remmina_file, "profile-lock", lock_enabled);
+	gboolean disconect_prompt = (priv->behavior_disconnect ? gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->behavior_disconnect)) : FALSE);
+	remmina_file_set_int(priv->remmina_file, "disconnect-prompt", disconect_prompt);
 }
 
-static void remmina_file_editor_save_ssh_tunnel_tab(RemminaFileEditor *gfe)
-{
-	TRACE_CALL(__func__);
-	RemminaFileEditorPriv *priv = gfe->priv;
-	gboolean ssh_tunnel_enabled;
-	int ssh_tunnel_auth;
-
-	ssh_tunnel_enabled = (priv->ssh_tunnel_enabled_check ? gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_enabled_check)) : FALSE);
-	remmina_file_set_int(priv->remmina_file,
-			     "ssh_tunnel_loopback",
-			     (priv->ssh_tunnel_loopback_check ? gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_loopback_check)) : FALSE));
-	remmina_file_set_int(priv->remmina_file, "ssh_tunnel_enabled", ssh_tunnel_enabled);
-	remmina_file_set_string(priv->remmina_file, "ssh_tunnel_auth",
-				remmina_public_combo_get_active_text(GTK_COMBO_BOX(priv->ssh_tunnel_auth_combo)));
-	remmina_file_set_string(priv->remmina_file, "ssh_tunnel_username",
-				(ssh_tunnel_enabled ? gtk_entry_get_text(GTK_ENTRY(priv->ssh_tunnel_username_entry)) : NULL));
-	remmina_file_set_string(
-		priv->remmina_file,
-		"ssh_tunnel_server",
-		(ssh_tunnel_enabled && priv->ssh_tunnel_server_entry && (priv->ssh_tunnel_server_custom_radio == NULL || gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->ssh_tunnel_server_custom_radio))) ? gtk_entry_get_text(GTK_ENTRY(priv->ssh_tunnel_server_entry)) : NULL));
-
-	ssh_tunnel_auth = gtk_combo_box_get_active(GTK_COMBO_BOX(priv->ssh_tunnel_auth_combo));
-
-	remmina_file_set_int(
-		priv->remmina_file,
-		"ssh_tunnel_auth",
-		ssh_tunnel_auth);
-	remmina_file_set_string(
-		priv->remmina_file,
-		"ssh_tunnel_privatekey",
-		(priv->ssh_tunnel_privatekey_chooser ? gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(priv->ssh_tunnel_privatekey_chooser)) : NULL));
-
-	remmina_file_set_string(
-		priv->remmina_file,
-		"ssh_tunnel_certfile",
-		(priv->ssh_tunnel_certfile_chooser ? gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(priv->ssh_tunnel_certfile_chooser)) : NULL));
-
-	remmina_file_set_string(
-		priv->remmina_file,
-		"ssh_tunnel_password",
-		(ssh_tunnel_enabled && (ssh_tunnel_auth == SSH_AUTH_PASSWORD)) ? gtk_entry_get_text(GTK_ENTRY(priv->ssh_tunnel_auth_password)) : NULL);
-
-	remmina_file_set_string(
-		priv->remmina_file,
-		"ssh_tunnel_passphrase",
-		(ssh_tunnel_enabled && (ssh_tunnel_auth == SSH_AUTH_PUBLICKEY || ssh_tunnel_auth == SSH_AUTH_AUTO_PUBLICKEY)) ? gtk_entry_get_text(GTK_ENTRY(priv->ssh_tunnel_passphrase)) : NULL);
-}
 
 static gboolean remmina_file_editor_validate_settings(RemminaFileEditor *	gfe,
 						      gchar *			setting_name_to_validate,
@@ -1485,7 +1797,6 @@ static gboolean remmina_file_editor_validate_settings(RemminaFileEditor *	gfe,
 
 	setting_iter = protocol_plugin->basic_settings;
 	if (setting_iter) {
-		// gboolean found = FALSE;
 		while (setting_iter->type != REMMINA_PROTOCOL_SETTING_TYPE_END) {
 			if (setting_iter->name == NULL) {
 				g_error("Internal error: a setting name in protocol plugin %s is "
@@ -1493,7 +1804,6 @@ static gboolean remmina_file_editor_validate_settings(RemminaFileEditor *	gfe,
 					protocol_plugin->name);
 			} else if ((gchar *)setting_name_to_validate) {
 				if (strcmp((gchar *)setting_name_to_validate, setting_iter->name) == 0) {
-					// found = TRUE;
 
 					gpointer validator_data = setting_iter->validator_data;
 					GCallback validator = setting_iter->validator;
@@ -1522,11 +1832,6 @@ static gboolean remmina_file_editor_validate_settings(RemminaFileEditor *	gfe,
 			}
 			setting_iter++;
 		}
-
-		// if (!found) {
-		// 	 TOO VERBOSE:
-		// 	 g_warning("%s is not a plugin setting!", setting_name_to_validate);
-		// }
 	}
 
 	return TRUE;
@@ -1549,10 +1854,14 @@ static GError *remmina_file_editor_update_settings(RemminaFileEditor *	gfe,
 
 	g_hash_table_iter_init(&iter, priv->setting_widgets);
 	while (g_hash_table_iter_next(&iter, &key, &widget)) {
+		
 		// We don't want to save or validate grayed-out settings.
-		if (!gtk_widget_get_sensitive(GTK_WIDGET(widget))) {
+		// If widget is a file chooser, it was made not sensitive because
+		// the box was unchecked. In that case, don't continue. The 
+		// relevant file strings will be set to NULL in the remmina file.
+		if (!gtk_widget_get_sensitive(GTK_WIDGET(widget)) && !GTK_IS_FILE_CHOOSER(widget)) {
 			g_debug("Grayed-out setting-widget '%s' will not be saved.",
-				gtk_widget_get_name(widget));
+			gtk_widget_get_name(widget));
 			continue;
 		}
 
@@ -1597,6 +1906,11 @@ static GError *remmina_file_editor_update_settings(RemminaFileEditor *	gfe,
 		} else if (GTK_IS_FILE_CHOOSER(widget)) {
 			gchar *value = gtk_widget_get_sensitive(GTK_WIDGET(widget)) ? gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(widget)) : NULL;
 
+			if (!gtk_widget_get_sensitive(GTK_WIDGET(widget))) {
+				remmina_file_set_string(priv->remmina_file, (gchar *)key, value);
+				continue;
+			}
+
 			if (!remmina_file_editor_validate_settings(gfe, (gchar *)key, value, &err)) {
 				// Error while validating!
 				// err should be set now.
@@ -1674,12 +1988,23 @@ static GError *remmina_file_editor_update(RemminaFileEditor *	gfe,
 		remmina_file_set_int(priv->remmina_file, "resolution_height", res_h);
 	}
 
+	if (priv->assistance_toggle){
+		if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->assistance_toggle))) {
+			remmina_file_set_string(priv->remmina_file, "assistance_file", gtk_entry_get_text(GTK_ENTRY(priv->assistance_file)));
+			remmina_file_set_string(priv->remmina_file, "assistance_pass", gtk_entry_get_text(GTK_ENTRY(priv->assistance_password)));
+			remmina_file_set_int(priv->remmina_file, "assistance_mode", 1);
+		}else{
+			remmina_file_set_int(priv->remmina_file, "assistance_mode", 0);
+		}
+		
+	}
+
 	if (priv->keymap_combo)
 		remmina_file_set_string(priv->remmina_file, "keymap",
 					remmina_public_combo_get_active_text(GTK_COMBO_BOX(priv->keymap_combo)));
 
 	remmina_file_editor_save_behavior_tab(gfe);
-	remmina_file_editor_save_ssh_tunnel_tab(gfe);
+	remmina_file_editor_save_ssh_tunnel_tab(gfe, NULL);
 	return remmina_file_editor_update_settings(gfe, failed_widget);
 }
 
@@ -1929,8 +2254,6 @@ GtkWidget *remmina_file_editor_new_from_file(RemminaFile *remminafile)
 	gtk_container_set_border_width(GTK_CONTAINER(grid), 8);
 	gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(GTK_DIALOG(gfe))), grid, FALSE, FALSE, 2);
 
-	// remmina_public_create_group(GTK_GRID(grid), _("Profile"), 0, 4, 3);
-
 	gboolean profile_file_exists = (remmina_file_get_filename(remminafile) != NULL);
 
 	/* Profile: Name */
